@@ -25,17 +25,21 @@
 #include <errno.h>
 #include <ctype.h>
 #include <sys/socket.h>
-#include <sys/select.h>
+#include <poll.h>
 #include <netinet/in.h>
 #include <netinet/icmp6.h>
 #include <arpa/inet.h>
 #include <net/ethernet.h>
 #include <assert.h>
 #include <ndp.h>
+#include <net/if.h>
+#include <sys/ioctl.h>
 
 #include "ndp_private.h"
 #include "list.h"
 
+#define pr_err(args...) fprintf(stderr, ##args)
+
 /**
  * SECTION: logging
  * @short_description: libndp logging facility
@@ -217,11 +221,9 @@ resend:
 	return 0;
 }
 
-static const char *str_in6_addr(struct in6_addr *addr)
+static const char *str_in6_addr(struct in6_addr *addr, char buf[static INET6_ADDRSTRLEN])
 {
-	static char buf[INET6_ADDRSTRLEN];
-
-	return inet_ntop(AF_INET6, addr, buf, sizeof(buf));
+	return inet_ntop(AF_INET6, addr, buf, INET6_ADDRSTRLEN);
 }
 
 
@@ -308,6 +310,18 @@ static void ndp_msg_addrto_adjust_all_routers(struct in6_addr *addr)
 	addr->s6_addr32[3] = htonl(0x2);
 }
 
+/*
+ * compute link-local solicited-node multicast address
+ */
+static void ndp_msg_addrto_adjust_solicit_multi(struct in6_addr *addr,
+						struct in6_addr *target)
+{
+	addr->s6_addr32[0] = htonl(0xFF020000);
+	addr->s6_addr32[1] = 0;
+	addr->s6_addr32[2] = htonl(0x1);
+	addr->s6_addr32[3] = htonl(0xFF000000) | target->s6_addr32[3];
+}
+
 static bool ndp_msg_addrto_validate_link_local(struct in6_addr *addr)
 {
 	return IN6_IS_ADDR_LINKLOCAL (addr);
@@ -679,6 +693,137 @@ void ndp_msg_ifindex_set(struct ndp_msg *msg, uint32_t ifindex)
 	msg->ifindex = ifindex;
 }
 
+/**
+ * ndp_msg_dest_set:
+ * @msg: message structure
+ * @dest: ns,na dest
+ *
+ * Set dest address in IPv6 header for NS and NA.
+ **/
+NDP_EXPORT
+void ndp_msg_dest_set(struct ndp_msg *msg, struct in6_addr *dest)
+{
+	enum ndp_msg_type msg_type = ndp_msg_type(msg);
+	switch (msg_type) {
+		case NDP_MSG_NS:
+			/* fall through */
+		case NDP_MSG_NA:
+			msg->addrto = *dest;
+			/* fall through */
+		default:
+			break;
+	}
+}
+
+/**
+ * ndp_msg_target_set:
+ * @msg: message structure
+ * @target: ns,na target
+ *
+ * Set target address in ICMPv6 header for NS and NA.
+ **/
+NDP_EXPORT
+void ndp_msg_target_set(struct ndp_msg *msg, struct in6_addr *target)
+{
+	struct in6_addr any = IN6ADDR_ANY_INIT;
+	enum ndp_msg_type msg_type = ndp_msg_type(msg);
+
+	switch (msg_type) {
+		case NDP_MSG_NS:
+			((struct ndp_msgns*)&msg->nd_msg)->ns->nd_ns_target = *target;
+			/*
+			 * Neighbor Solicitations are multicast when the node
+			 * needs to resolve an address and unicast when the
+			 * node seeks to verify the reachability of a
+			 * neighbor.
+			 *
+			 * In this case we need to update the dest address in
+			 * IPv6 header when
+			 * a) IPv6 dest address is not set
+			 * b) ICMPv6 target address is supplied
+			 * */
+			if (!memcmp(&msg->addrto, &any, sizeof(any)) &&
+			    memcmp(target, &any, sizeof(any)))
+				ndp_msg_addrto_adjust_solicit_multi(&msg->addrto, target);
+			break;
+		case NDP_MSG_NA:
+			((struct ndp_msgna*)&msg->nd_msg)->na->nd_na_target = *target;
+			break;
+		default:
+			break;
+	}
+}
+
+static int ndp_get_iface_mac(int ifindex, char *ptr)
+{
+	int sockfd, err = 0;
+	struct ifreq ifr;
+
+	sockfd = socket(AF_INET, SOCK_DGRAM, 0);
+	if (sockfd == -1) {
+		pr_err("%s: Failed to create socket", __func__);
+		return -errno;
+	}
+
+	if (if_indextoname(ifindex, (char *)&ifr.ifr_name) == NULL) {
+		pr_err("%s: Failed to get iface name with index %d", __func__, ifindex);
+		err = -errno;
+		goto close_sock;
+	}
+
+	if (ioctl(sockfd, SIOCGIFHWADDR, &ifr) < 0) {
+		pr_err("%s: Failed to get iface mac with index %d\n", __func__, ifindex);
+		err = -errno;
+		goto close_sock;
+	}
+
+	memcpy(ptr, &ifr.ifr_hwaddr.sa_data, sizeof(ifr.ifr_hwaddr.sa_data));
+
+close_sock:
+	close(sockfd);
+	return err;
+}
+
+static void ndp_msg_opt_set_linkaddr(struct ndp_msg *msg, int ndp_opt)
+{
+	char *opts_start = ndp_msg_payload_opts(msg);
+	struct nd_opt_hdr *s_laddr_opt = (struct nd_opt_hdr *) opts_start;
+	char *opt_data = (char *) s_laddr_opt + sizeof(struct nd_opt_hdr);
+	int err;
+
+	err = ndp_get_iface_mac(ndp_msg_ifindex(msg), opt_data);
+	if (err)
+		return;
+
+	opt_data += 6;
+	s_laddr_opt->nd_opt_type = ndp_opt;
+	s_laddr_opt->nd_opt_len = (opt_data - opts_start) >> 3;
+	msg->len += opt_data - opts_start;
+}
+
+/**
+ * ndp_msg_opt_set:
+ * @msg: message structure
+ *
+ * Set neighbor discovery option info.
+ **/
+NDP_EXPORT
+void ndp_msg_opt_set(struct ndp_msg *msg)
+{
+	enum ndp_msg_type msg_type = ndp_msg_type(msg);
+
+	switch (msg_type) {
+		case NDP_MSG_NS:
+			ndp_msg_opt_set_linkaddr(msg, ND_OPT_SOURCE_LINKADDR);
+			break;
+		case NDP_MSG_NA:
+			ndp_msg_opt_set_linkaddr(msg, ND_OPT_TARGET_LINKADDR);
+			break;
+		default:
+			break;
+	}
+}
+
 /**
  * ndp_msg_send:
  * @ndp: libndp library context
@@ -1086,6 +1231,37 @@ static bool ndp_msg_opt_route_check_valid(void *opt_data)
 	 */
 	if (((ri->nd_opt_ri_prf_reserved >> 3) & 3) == 2)
 		return false;
+
+	/* The Length field is 1, 2, or 3 depending on the Prefix Length.
+	 * If Prefix Length is greater than 64, then Length must be 3.
+	 * If Prefix Length is greater than 0, then Length must be 2 or 3.
+	 * If Prefix Length is zero, then Length must be 1, 2, or 3.
+	 */
+	if (ri->nd_opt_ri_len > 3 ||
+	    (ri->nd_opt_ri_prefix_len > 64 && ri->nd_opt_ri_len != 3) ||
+	    (ri->nd_opt_ri_prefix_len > 0 && ri->nd_opt_ri_len == 1))
+		return false;
+
+	return true;
+}
+
+static bool ndp_msg_opt_pref64_check_valid(void *opt_data)
+{
+	struct __nd_opt_pref64 *pref64 = opt_data;
+	uint16_t plc;
+
+	/* rfc8781: the receiver MUST ignore the PREF64 option if the
+	 * Length field value is not 2 */
+	if (pref64->nd_opt_pref64_len != 2)
+		return false;
+
+	/* The PLC field values 0, 1, 2, 3, 4, and 5 indicate the NAT64
+	 * prefix length. The receiver MUST ignore the PREF64 option if the
+	 * Prefix Length Code field is not set to one of those values.*/
+	plc = ntohs(pref64->nd_opt_pref64_lft_plc) & ND_OPT_PREF64_PLC_MASK;
+	if (plc > 5)
+		return false;
+
 	return true;
 }
 
@@ -1121,6 +1297,11 @@ static struct ndp_msg_opt_type_info ndp_msg_opt_type_info_list[] =
 		.raw_type = __ND_OPT_DNSSL,
 		.raw_struct_size = sizeof(struct __nd_opt_dnssl),
 	},
+	[NDP_MSG_OPT_PREF64] = {
+		.raw_type = __ND_OPT_PREF64,
+		.raw_struct_size = sizeof(struct __nd_opt_pref64),
+		.check_valid = ndp_msg_opt_pref64_check_valid,
+	},
 };
 
 #define NDP_MSG_OPT_TYPE_LIST_SIZE ARRAY_SIZE(ndp_msg_opt_type_info_list)
@@ -1468,7 +1649,7 @@ uint32_t ndp_msg_opt_mtu(struct ndp_msg *msg, int offset)
 NDP_EXPORT
 struct in6_addr *ndp_msg_opt_route_prefix(struct ndp_msg *msg, int offset)
 {
-	static struct in6_addr prefix;
+	static NDP_THREAD struct in6_addr prefix;
 	struct __nd_opt_route_info *ri =
 			ndp_msg_payload_opts_offset(msg, offset);
 
@@ -1574,7 +1755,7 @@ NDP_EXPORT
 struct in6_addr *ndp_msg_opt_rdnss_addr(struct ndp_msg *msg, int offset,
 					int addr_index)
 {
-	static struct in6_addr addr;
+	static NDP_THREAD struct in6_addr addr;
 	struct __nd_opt_rdnss *rdnss =
 			ndp_msg_payload_opts_offset(msg, offset);
 	size_t len = rdnss->nd_opt_rdnss_len << 3; /* convert to bytes */
@@ -1624,7 +1805,7 @@ char *ndp_msg_opt_dnssl_domain(struct ndp_msg *msg, int offset,
 			       int domain_index)
 {
 	int i;
-	static char buf[256];
+	static NDP_THREAD char buf[256];
 	struct __nd_opt_dnssl *dnssl =
 			ndp_msg_payload_opts_offset(msg, offset);
 	size_t len = dnssl->nd_opt_dnssl_len << 3; /* convert to bytes */
@@ -1665,6 +1846,86 @@ char *ndp_msg_opt_dnssl_domain(struct ndp_msg *msg, int offset,
 	return NULL;
 }
 
+/**
+ * ndp_msg_opt_pref64_lifetime:
+ * @msg: message structure
+ * @offset: in-message offset
+ *
+ * Get the NAT64 prefix lifetime.
+ * User should use this function only inside ndp_msg_opt_for_each_offset()
+ * macro loop.
+ *
+ * Returns: the NAT64 prefix lifetime in seconds.
+ **/
+NDP_EXPORT
+uint16_t ndp_msg_opt_pref64_lifetime(struct ndp_msg *msg, int offset)
+{
+	struct __nd_opt_pref64 *pref64 =
+			ndp_msg_payload_opts_offset(msg, offset);
+
+	return ntohs(pref64->nd_opt_pref64_lft_plc) & ND_OPT_PREF64_LFT_MASK;
+}
+
+static uint8_t ndp_msg_opt_pref64_plen(struct __nd_opt_pref64 *pref64)
+{
+	uint8_t plc;
+
+	plc = ntohs(pref64->nd_opt_pref64_lft_plc) & ND_OPT_PREF64_PLC_MASK;
+
+	/* The PLC field values 0, 1, 2, 3, 4, and 5 indicate the NAT64
+	 * prefix length of 96, 64, 56, 48, 40, and 32 bits, respectively. */
+	if (plc == 0)
+		return 96;
+
+	return 72 - plc * 8;
+}
+
+/**
+ * ndp_msg_opt_pref64_prefix_length:
+ * @msg: message structure
+ * @offset: in-message offset
+ *
+ * Get the NAT64 prefix length.
+ * User should use this function only inside ndp_msg_opt_for_each_offset()
+ * macro loop.
+ *
+ * Returns: the NAT64 prefix length (32, 40, 48, 56, 64 or 96).
+ **/
+NDP_EXPORT
+uint8_t ndp_msg_opt_pref64_prefix_length(struct ndp_msg *msg, int offset)
+{
+	struct __nd_opt_pref64 *pref64 =
+			ndp_msg_payload_opts_offset(msg, offset);
+
+	return ndp_msg_opt_pref64_plen(pref64);
+}
+
+/**
+ * ndp_msg_opt_pref64_prefix:
+ * @msg: message structure
+ * @offset: in-message offset
+ *
+ * Get the NAT64 prefix.
+ * User should use this function only inside ndp_msg_opt_for_each_offset()
+ * macro loop.
+ *
+ * Returns: the NAT64 prefix.
+ **/
+NDP_EXPORT
+struct in6_addr *ndp_msg_opt_pref64_prefix(struct ndp_msg *msg, int offset)
+{
+	static NDP_THREAD struct in6_addr addr;
+	struct __nd_opt_pref64 *pref64 =
+			ndp_msg_payload_opts_offset(msg, offset);
+	uint8_t plen;
+
+	plen = ndp_msg_opt_pref64_plen(pref64);
+	memset(&addr, 0, sizeof(addr));
+	memcpy(&addr, pref64->nd_opt_pref64_prefix, plen / 8);
+
+	return &addr;
+}
+
 static int ndp_call_handlers(struct ndp *ndp, struct ndp_msg *msg);
 
 static int ndp_sock_recv(struct ndp *ndp)
@@ -1673,6 +1934,7 @@ static int ndp_sock_recv(struct ndp *ndp)
 	enum ndp_msg_type msg_type;
 	size_t len;
 	int err;
+	char buf[INET6_ADDRSTRLEN];
 
 	msg = ndp_msg_alloc();
 	if (!msg)
@@ -1686,7 +1948,7 @@ static int ndp_sock_recv(struct ndp *ndp)
 		goto free_msg;
 	}
 	dbg(ndp, "rcvd from: %s, ifindex: %u, hoplimit: %d",
-		 str_in6_addr(&msg->addrto), msg->ifindex, msg->hoplimit);
+		 str_in6_addr(&msg->addrto, buf), msg->ifindex, msg->hoplimit);
 
 	if (msg->hoplimit != 255) {
 		warn(ndp, "ignoring packet with bad hop limit (%d)", msg->hoplimit);
@@ -1961,22 +2223,20 @@ int ndp_call_eventfd_handler(struct ndp *ndp)
 NDP_EXPORT
 int ndp_callall_eventfd_handler(struct ndp *ndp)
 {
-	fd_set rfds;
-	int fdmax;
-	struct timeval tv;
-	int fd = ndp_get_eventfd(ndp);
+	struct pollfd pfd;
 	int ret;
 	int err;
 
-	memset(&tv, 0, sizeof(tv));
-	FD_ZERO(&rfds);
-	FD_SET(fd, &rfds);
-	fdmax = fd + 1;
+	pfd = (struct pollfd) {
+		.fd = ndp_get_eventfd(ndp),
+		.events = POLLIN,
+	};
+
 	while (true) {
-		ret = select(fdmax, &rfds, NULL, NULL, &tv);
+		ret = poll(&pfd, 1, 0);
 		if (ret == -1)
 			return -errno;
-		if (!FD_ISSET(fd, &rfds))
+		if (!(pfd.revents & POLLIN))
 			return 0;
 		err = ndp_call_eventfd_handler(ndp);
 		if (err)
